/*  Copyright (C) 1996-1997  Id Software, Inc.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

    See file, 'COPYING', for details.
*/

/*
 * tyrlite/entities.c
 * Modifications by Kevin Shanahan, 1999-2000
 */

#include <string.h>
#include "cmdlib.h"

#include "tyrlite.h"
#include "entities.h"

entity_t entities[MAX_MAP_ENTITIES];
int      num_entities;

/*
 * ============================================================================
 * ENTITY FILE PARSING
 * If a light has a targetname, generate a unique style in the 32-63 range
 * ============================================================================
 */
 
#define MAX_LIGHT_TARGETS 32
 
int  numlighttargets;
char lighttargets[MAX_LIGHT_TARGETS][MAX_ENT_VALUE];

static void SetKeyValue (entity_t *, const char *, const char *);

static int LightStyleForTargetname (const char *targetname, qboolean alloc)
{
    int i;

    for (i=0 ; i<numlighttargets ; i++)
        if (!strcmp (lighttargets[i], targetname))
            return 32 + i;
    if (!alloc)
        return -1;
        
    if (i == MAX_LIGHT_TARGETS)
		Error("LightStyleForTargetname: Too many unique light targetnames");
	
    strcpy (lighttargets[i], targetname);
    numlighttargets++;

    return numlighttargets-1 + 32;
}

/*
 * ==================
 * MatchTargets
 * ==================
 */
static void MatchTargets (void)
{
    int i;
    int j;

    for (i=0 ; i<num_entities ; i++) {
        if (!entities[i].target[0])
            continue;
        for (j=0 ; j<num_entities ; j++)
            if (!strcmp(entities[j].targetname, entities[i].target)) {
                entities[i].targetent = &entities[j];
                break;
            }
        if (j==num_entities) {
            printf("WARNING: entity at (%i,%i,%i) (%s) has unmatched target\n",
		   (int)entities[i].origin[0], (int)entities[i].origin[1],
		   (int)entities[i].origin[2], entities[i].classname);
            continue;
        }

        /* set the style on the source ent for switchable lights */
        if (entities[j].style) {
            char    s[16];
            
            entities[i].style = entities[j].style;
            sprintf (s,"%i", entities[i].style);
            SetKeyValue (&entities[i], "style", s);
        }
    }    
}


/*
 * ==================
 * LoadEntities
 * ==================
 */
void LoadEntities (void)
{
    char     *data;
    entity_t *entity;
    char     key[MAX_ENT_KEY];
    epair_t  *epair;
    double   vec[3];
    int      i;
    int      num_lights;

    data = dentdata;

    /* start parsing */
    num_entities = 0;
    num_lights = 0;

    /* go through all the entities */
    while (1) {
	/* parse the opening brace */
        data = COM_Parse (data);
        if (!data)
            break;
        if (com_token[0] != '{')
            Error ("LoadEntities: found %s when expecting {",com_token);

        if (num_entities == MAX_MAP_ENTITIES)
            Error ("LoadEntities: MAX_MAP_ENTITIES");
        entity = &entities[num_entities];
        num_entities++;

	/* go through all the keys in this entity */
        while (1) {
            int c;

	    /* parse key */
            data = COM_Parse (data);
            if (!data)
                Error ("LoadEntities: EOF without closing brace");
            if (!strcmp(com_token,"}"))
                break;
           	if (strlen(com_token) >= MAX_ENT_KEY)
				Error ("LoadEntities: Key length > %i", MAX_ENT_KEY - 1);
            strcpy (key, com_token);

	    /* parse value */
            data = COM_Parse (data);
            if (!data)
                Error ("LoadEntities: EOF without closing brace");
            c = com_token[0];
            if (c == '}')
                Error ("LoadEntities: closing brace without data");
			if (strlen(com_token) >= MAX_ENT_VALUE)
				Error("LoadEntities: Value length > %i", MAX_ENT_VALUE - 1);
		
            epair = malloc (sizeof(epair_t));
            memset (epair, 0, sizeof(epair));
            strcpy (epair->key, key);
            strcpy (epair->value, com_token);
            epair->next = entity->epairs;
            entity->epairs = epair;

            if (!strcmp(key, "classname"))
                strcpy (entity->classname, com_token);
            else if (!strcmp(key, "target"))
                strcpy (entity->target, com_token);
            else if (!strcmp(key, "targetname"))
                strcpy (entity->targetname, com_token);
            else if (!strcmp(key, "origin")) {
                /* scan into doubles, then assign        */
		/* which makes it vec_t size independent */
                if (sscanf(com_token, "%lf %lf %lf",
			   &vec[0], &vec[1], &vec[2]) != 3)
                    Error ("LoadEntities: not 3 values for origin");
                for (i=0 ; i<3 ; i++)
                    entity->origin[i] = vec[i];
            } else if (!strncmp(key, "light", 5) || !strcmp (key, "_light"))
                entity->light = atof(com_token);
            else if (!strcmp(key, "style")) {
                entity->style = atof(com_token);
                if ((unsigned)entity->style > 254)
                    Error("Bad light style %i (must be 0-254)", entity->style);
            } else if (!strcmp(key, "angle"))
                entity->angle = atof(com_token);
            else if (!strcmp(key, "wait"))
                entity->atten = atof(com_token);
            else if (!strcmp(key, "delay"))
                entity->formula = atoi(com_token);
            else if (!strcmp(key, "mangle")) {
                /* scan into doubles, then assign        */
		/* which makes it vec_t size independent */
                if (sscanf(com_token, "%lf %lf %lf",
			   &vec[0], &vec[1], &vec[2]) != 3)
                    Error ("LoadEntities: not 3 values for mangle");

                /* Precalculate the direction vector */
                entity->use_mangle = true;
                entity->mangle[0] = cos(vec[0]*Q_PI/180)*cos(vec[1]*Q_PI/180);
                entity->mangle[1] = sin(vec[0]*Q_PI/180)*cos(vec[1]*Q_PI/180);
                entity->mangle[2] = sin(vec[1]*Q_PI/180);
            } else if (!strcmp(key, "_color") || !strcmp(key, "color")) {
                /* scan into doubles, then assign        */
                /* which makes it vec_t size independent */
                if (sscanf(com_token, "%lf %lf %lf",
 				&vec[0], &vec[1], &vec[2]) != 3) {
                    printf("WARNING - LoadEntities: "
			     "not 3 values for color\n");
			} else {
                for (i=0 ; i<3 ; i++)
                    entity->lightcolour[i] = vec[i];
          	}
            } else if (!strcmp(key, "_sunlight"))
                sunlight = atof(com_token);
            else if (!strcmp(key, "_sun_mangle")) {

                /* scan into doubles, then assign        */
                /* which makes it vec_t size independent */
                if (sscanf(com_token, "%lf %lf %lf",
			   &vec[0], &vec[1], &vec[2]) != 3)
                    Error ("LoadEntities: not 3 values for _sun_mangle");
                
                /* Precalculate sun vector and           */
                /* make it too large to fit into the map */
                sunmangle[0] = cos(vec[0]*Q_PI/180)*cos(vec[1]*Q_PI/180);
                sunmangle[1] = sin(vec[0]*Q_PI/180)*cos(vec[1]*Q_PI/180);
                sunmangle[2] = sin(vec[1]*Q_PI/180);
                VectorNormalize(sunmangle);
                VectorScale(sunmangle, -16384, sunmangle);
            } else if (!strcmp(key, "_sunlight_color")) {
                /* scan into doubles, then assign        */
                /* which makes it vec_t size independent */
                if (sscanf(com_token, "%lf %lf %lf",
			   &vec[0], &vec[1], &vec[2]) != 3)
                    Error ("LoadEntities: not 3 values for _sunlight_color");
                for (i=0 ; i<3 ; i++)
                    sunlight_color[i] = vec[i];
            } else if (!strcmp(key, "_minlight_color")) {
                /* scan into doubles, then assign        */
                /* which makes it vec_t size independent */
                if (sscanf(com_token, "%lf %lf %lf",
			   &vec[0], &vec[1], &vec[2]) != 3)
                    Error ("LoadEntities: not 3 values for _minlight_color");
                for (i=0 ; i<3 ; i++)
                    minlight_color[i] = vec[i];
            }
            // pOx - START - Noise from Open Quartz
            else if (!strcmp(key, "_noise"))
				entity->noise = atof (com_token);
			else if (!strcmp(key, "_noisetype"))
			{
				// pOx - Quiver has no string-select (but shouldn't hurt to leave in)
				if (!strcmp(com_token, "random"))
					entity->noisetype = NOISE_RANDOM;
				else if (!strcmp(com_token, "smooth"))
					entity->noisetype = NOISE_SMOOTH;
				else if (!strcmp(com_token, "perlin"))
					entity->noisetype = NOISE_PERLIN;
				else
					entity->noisetype = atoi (com_token);
			}
			else if (!strcmp(key, "_persistence"))
				entity->persistence = atof (com_token);
			else if (!strcmp(key, "_resolution"))
				entity->resolution = atof (com_token);
			// pOx - END - Noise from Open Quartz
			else if (!strcmp(key, "_notents"))// pOx - this light doesn't affect sub models
       			if (atoi(com_token)) entity->notents = true;
        }

	/* All fields have been parsed. Check default settings */
        /* and check for light value in worldspawn...          */
        if (!strncmp (entity->classname, "light", 5)) {
             // pOx - START - Noise from Open Quartz
            if (!entity->noise) entity->noise = noise;
			if (!entity->persistence) entity->persistence = 1;
			if (!entity->resolution) entity->resolution = 1;// pOx - added safe default
			// pOx - END - Noise from Open Quartz 
            if (!entity->light)
                entity->light = DEFAULTLIGHTLEVEL;
            if (entity->atten <= 0.0)
                entity->atten = 1.0;
            if ((entity->formula < 0) || (entity->formula > 3))
                entity->formula = 0;
            if (!entity->lightcolour[0] && !entity->lightcolour[1] && !entity->lightcolour[2]) {
				entity->lightcolour[0] = 255;
                entity->lightcolour[1] = 255;
                entity->lightcolour[2] = 255;
            }
            num_lights++;
        }

        if (!strcmp (entity->classname, "light")) {
            if (entity->targetname[0] && !entity->style) {
                char s[16];

                entity->style =
		    LightStyleForTargetname (entity->targetname, true);
                sprintf (s,"%i", entity->style);
                SetKeyValue (entity, "style", s);
            }
        }

        if (!strcmp (entity->classname, "worldspawn")) {
            if (entity->light > 0 && !worldminlight) {
                worldminlight = entity->light;
                printf("using minlight value %i from worldspawn.\n",
		       worldminlight);
            } else if (worldminlight) {
                printf("Using minlight value %i from command line.\n",
		       worldminlight);
            }
        }
    }
    printf ("%d entities read, %d are lights.\n", num_entities, num_lights);
    MatchTargets ();
}

char *ValueForKey (const entity_t *ent, const char *key)
{
    epair_t *ep;

    for (ep=ent->epairs ; ep ; ep=ep->next)
        if (!strcmp (ep->key, key))
            return ep->value;
    return "";
}

static void SetKeyValue (entity_t *ent, const char *key, const char *value)
{
    epair_t *ep;

    for (ep=ent->epairs ; ep ; ep=ep->next)
        if (!strcmp (ep->key, key) ) {
            strcpy (ep->value, value);
            return;
        }
    ep = malloc (sizeof(*ep));
    ep->next = ent->epairs;
    ent->epairs = ep;
    strcpy (ep->key, key);
    strcpy (ep->value, value);
}

entity_t *FindEntityWithKeyPair( const char *key, const char *value )
{
    entity_t *ent;
    epair_t  *ep;
    int i;

    for (i=0 ; i<num_entities ; i++) {
        ent = &entities[ i ];
        for (ep=ent->epairs ; ep ; ep=ep->next)
            if (!strcmp (ep->key, key)) {
                if (!strcmp( ep->value, value ))
		    return ent;
                break;
            }
    }
    return NULL;
}

/*
float FloatForKey (entity_t *ent, char *key)
{
    char *k;

    k = ValueForKey (ent, key);
    return atof(k);
}
*/
void GetVectorForKey (const entity_t *ent, const char *key, vec3_t vec)
{
    char *k;

    k = ValueForKey (ent, key);
    sscanf (k, "%f %f %f", &vec[0], &vec[1], &vec[2]);
}


/*
 * ================
 * WriteEntitiesToString
 * ================
 */
void WriteEntitiesToString (void)
{
    char    *buf;
    char    *end;
    epair_t *ep;
    int     i;
	char    line[MAX_ENT_KEY + MAX_ENT_VALUE + 6];
    /* Room for 4 double quotes, a space and trailing null. */
    
    buf = dentdata;
    end = buf;
    *end = 0;

    printf ("%i switchable light styles\n", numlighttargets);

    for (i=0 ; i<num_entities ; i++) {
        ep = entities[i].epairs;
        if (!ep)
            continue;    /* ent got removed */

        strcat (end,"{\n");
        end += 2;

        for (ep = entities[i].epairs ; ep ; ep=ep->next) {
          if (compress_ents && !strncmp(entities[i].classname, "light", 5)) {
                if (!strcmp(ep->key, "classname")  ||
		    !strcmp(ep->key, "origin")     ||
		    !strcmp(ep->key, "targetname") ||
		    !strcmp(ep->key, "spawnflags")) {
                    sprintf (line, "\"%s\" \"%s\"\n", ep->key, ep->value);
                    strcat (end, line);
                    end += strlen(line);
                }
            } else {
                sprintf (line, "\"%s\" \"%s\"\n", ep->key, ep->value);
                strcat (end, line);
                end += strlen(line);
            }
        }

        strcat (end,"}\n");
        end += 2;

        if (end > buf + MAX_MAP_ENTSTRING)
            Error ("Entity text too long");
    }
    entdatasize = end - buf + 1;
}
