/*  Copyright (C) 1996-1997  Id Software, Inc.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

    See file, 'COPYING', for details.
*/

/*
 * tyrlite/ltface.c
 * Modifications by Kevin Shanahan, 1999-2000
 */

#include "tyrlite.h"

#if TYRLIGHT_TEST_PHONG
#include "phong.h"
// DEBUG
int modnorms = 0;
int facenorms = 0;
#endif

// pOx - START - Noise from Open Quartz
/*
============
noise3d

returns the 3D noise value for a point in space
============
*/
static vec_t noise3d (vec3_t v, int num)
{
	int n;
	
	num;// pOx - shut up compiler
	
	n = floor(v[0]) + floor(v[1]) * 57 + floor(v[2]) * 3251;
	n = (n << 13) ^ n;
	n = n * (n * n * 15731 + 789221) + 1376312589;
	return (n & 0x7fffffff) / 2147483648.0;
}

/*
===========
noiseXYZ

a variation of Noise3D that takes 3 floats
===========
*/
static vec_t noiseXYZ (float x, float y, float z, int num)
{
	vec3_t v;
	v[0] = x;
	v[1] = y;
	v[2] = z;
	return noise3d (v, num);
}

/*
============
noise_scaled

returns a noise value from a scaled up noise pattern

Actually created by sampling points on a certain grid 
and interpolating the inbetween value, making the noise smooth
============
*/
static vec_t noise_scaled (vec3_t v, float s, int num)
{
	float n = 0;
	vec3_t a,b,c,d;

	a[0]=floor(v[0]/s)*s;  b[0]=a[0]+s;  c[0]=(v[0]-a[0])/s;  d[0]=1-c[0];
	a[1]=floor(v[1]/s)*s;  b[1]=a[1]+s;  c[1]=(v[1]-a[1])/s;  d[1]=1-c[1];
	a[2]=floor(v[2]/s)*s;  b[2]=a[2]+s;  c[2]=(v[2]-a[2])/s;  d[2]=1-c[2];

	n += noiseXYZ (a[0], a[1], a[2], num) * d[0] * d[1] * d[2];
	n += noiseXYZ (a[0], a[1], b[2], num) * d[0] * d[1] * c[2];
	n += noiseXYZ (a[0], b[1], a[2], num) * d[0] * c[1] * d[2];
	n += noiseXYZ (a[0], b[1], b[2], num) * d[0] * c[1] * c[2];
	n += noiseXYZ (b[0], a[1], a[2], num) * c[0] * d[1] * d[2];
	n += noiseXYZ (b[0], a[1], b[2], num) * c[0] * d[1] * c[2];
	n += noiseXYZ (b[0], b[1], a[2], num) * c[0] * c[1] * d[2];
	n += noiseXYZ (b[0], b[1], b[2], num) * c[0] * c[1] * c[2];

	return n;
}

/*
============
noise_perlin

returns a Perlin noise value for a point in 3D space\

This noise function combines noise at several different scales
which makes it slower than just using one layer of noise
============
*/
static vec_t noise_perlin (vec3_t v, float p, int num)
{
	float n = 0;

	n +=        noise_scaled (v, 1024, num);
	n += 0.5 + (noise_scaled (v,  256, num) - 0.5) * p;
	n += 0.5 + (noise_scaled (v,   64, num) - 0.5) * p * p;
	n += 0.5 + (noise_scaled (v,   16, num) - 0.5) * p * p * p;

	return n / 2 - 0.5;
}

/*
===========
snap_vector

Use to create low-res noise patterns without interpolation
A good strategy for avoiding seams on terrain and curves
when called with a scale value of about 32

This is because surfaces have some texture pixels which are
actually not in the polygon at all so their edges don't
perfectly fit together.

Creates a random pattern of light and dark squares on flat
axis aligned surfaces, which is usually not what you want
but it may have interesting applications
===========
*/
void snap_vector (vec3_t v_old, vec3_t v_new, float scale)
{
	if (scale <= 0)
	{
		v_new[0] = v_old[0];
		v_new[1] = v_old[1];
		v_new[2] = v_old[2];
	}
	else
	{
		v_new[0] = floor (v_old[0] / scale + 0.5) * scale;
		v_new[1] = floor (v_old[1] / scale + 0.5) * scale;
		v_new[2] = floor (v_old[2] / scale + 0.5) * scale;
	}
}
// pOx - END - Noise from Open Quartz

// Solve three simultaneous equations
// mtx is modified by the function...
#define ZERO_EPSILON (0.001)
static qboolean
LU_Decompose (vec3_t mtx[3], int r[3], int c[2])
{
    int         i, j, k;    // loop variables
    vec_t       max;
    int         max_r, max_c;

    // Do gauss elimination
    for (i = 0; i < 3; ++i) {
	max = 0;
	max_r = max_c = i;
	for (j = i; j < 3; ++j) {
	    for (k = i; k < 3; ++k) {
		if (fabs (mtx[j][k]) > max) {
		    max = fabs (mtx[j][k]);
		    max_r = j;
		    max_c = k;
		}
	    }
	}

	// Check for parallel planes
	if (max < ZERO_EPSILON)
	    return false;

	// Swap rows/columns if necessary
	if (max_r != i) {
	    for (j = 0; j < 3; ++j) {
		max = mtx[i][j];
		mtx[i][j] = mtx[max_r][j];
		mtx[max_r][j] = max;
	    }
	    k = r[i];
	    r[i] = r[max_r];
	    r[max_r] = k;
	}
	if (max_c != i) {
	    for (j = 0; j < 3; ++j) {
		max = mtx[j][i];
		mtx[j][i] = mtx[j][max_c];
		mtx[j][max_c] = max;
	    }
	    k = c[i];
	    c[i] = c[max_c];
	    c[max_c] = k;
	}

	// Do pivot
	for (j = i + 1; j < 3; ++j) {
	    mtx[j][i] /= mtx[i][i];
	    for (k = i + 1; k < 3; ++k)
		mtx[j][k] -= mtx[j][i] * mtx[i][k];
	}
    }

    return true;
}

/* FIXME - comments... */
static void
solve3 (const vec3_t mtx[3], const int r[3], const int c[3],
	const vec3_t rhs, vec3_t soln)
{
    vec3_t y;

    // forward-substitution
    y[0] = rhs[r[0]];
    y[1] = rhs[r[1]] - mtx[1][0] * y[0];
    y[2] = rhs[r[2]] - mtx[2][0] * y[0] - mtx[2][1] * y[1];

    // back-substitution
    soln[c[2]] = y[2] / mtx[2][2];
    soln[c[1]] = (y[1] - mtx[1][2] * soln[c[2]]) / mtx[1][1];
    soln[c[0]] = (y[0] - mtx[0][1] * soln[c[1]] - mtx[0][2] * soln[c[2]])
	/ mtx[0][0];
}

/*
 * ============
 * CastRay
 * Returns the distance between the points, or -1 if blocked
 * =============
 */
static vec_t CastRay (const vec3_t p1, const vec3_t p2)
{
    int      i;
    vec_t    t;
    qboolean trace;

    trace = TestLine (p1, p2);

    if (!trace)
        return -1;        /* ray was blocked */

    t = 0;
    for (i=0 ; i< 3 ; i++)
        t += (p2[i]-p1[i]) * (p2[i]-p1[i]);

    if (t == 0)
        t = 1;            /* don't blow up... */
    return sqrt(t);
}

/*
 * ============================================================================
 * SAMPLE POINT DETERMINATION
 * void SetupBlock (dface_t *f) Returns with surfpt[] set
 *
 * This is a little tricky because the lightmap covers more area than the face.
 * If done in the straightforward fashion, some of the sample points will be
 * inside walls or on the other side of walls, causing false shadows and light
 * bleeds.
 *
 * To solve this, I only consider a sample point valid if a line can be drawn
 * between it and the exact midpoint of the face.  If invalid, it is adjusted
 * towards the center until it is valid.
 *
 * (this doesn't completely work)
 * ============================================================================
 */

#define SINGLEMAP (18*18*4)

typedef struct {
    vec_t   lightmaps[MAXLIGHTMAPS][SINGLEMAP];
    int     numlightstyles;
    vec_t   *light;
    vec_t   facedist;
    vec3_t  facenormal;

    int     numsurfpt;
    vec3_t  surfpt[SINGLEMAP];

    /* FIXME - Comment this properly */
    vec_t   worldtotex[2][4]; // Copy of face->texinfo->vecs
    vec3_t  LU[3];
    int     row_p[3];
    int     col_p[3];

    vec_t   exactmins[2], exactmaxs[2];

    int     texmins[2], texsize[2];
    int     lightstyles[256];
    int     surfnum;
    dface_t *face;

    /* Colored lighting */
    vec3_t  lightmapcolours[MAXLIGHTMAPS][SINGLEMAP];
} lightinfo_t;


/*
 * ================
 * CalcFaceVectors
 * Fills in texorg, worldtotex. and textoworld
 * ================
 */
void CalcFaceVectors (lightinfo_t *l)
{
    texinfo_t *tex;
    int       i, j;

    /* convert from float to vec_t */
    tex = &texinfo[l->face->texinfo];
    for (i=0 ; i<2 ; i++)
        for (j=0 ; j<4 ; j++)
            l->worldtotex[i][j] = tex->vecs[i][j];

    /* Prepare LU and row, column permutations */
    for (i = 0; i < 3; ++i)
	l->row_p[i] = l->col_p[i] = i;
    VectorCopy(l->worldtotex[0], l->LU[0]);
    VectorCopy(l->worldtotex[1], l->LU[1]);
    VectorCopy(l->facenormal, l->LU[2]);

    /* Decompose the matrix. If we can't, texture axes are invalid. */
    if (!LU_Decompose(l->LU, l->row_p, l->col_p)) {
        Error ("Bad texture axes on face:\n"
               "   face point at (%5.3f, %5.3f, %5.3f)\n",
               dvertexes[ dedges[ l->face->firstedge ].v[0] ].point[0],
               dvertexes[ dedges[ l->face->firstedge ].v[0] ].point[1],
               dvertexes[ dedges[ l->face->firstedge ].v[0] ].point[2]);
    }
}

static void
tex_to_world(vec_t s, vec_t t, const lightinfo_t *l, vec3_t world)
{
    vec3_t rhs;
    rhs[0] = s - l->worldtotex[0][3];
    rhs[1] = t - l->worldtotex[1][3];
    rhs[2] = l->facedist + 1;         // one "unit" in front of surface

    solve3(l->LU, l->row_p, l->col_p, rhs, world);
}

/*
 * ================
 * CalcFaceExtents
 * Fills in s->texmins[] and s->texsize[]
 * also sets exactmins[] and exactmaxs[]
 * ================
 */
void CalcFaceExtents (lightinfo_t *l, const vec3_t faceoffset)
{
    dface_t   *s;
    vec_t     mins[2], maxs[2], val;
    int       i,j, e;
    dvertex_t *v;
    texinfo_t *tex;

    s = l->face;

    mins[0] = mins[1] = 999999;
    maxs[0] = maxs[1] = -99999;

    tex = &texinfo[s->texinfo];

    for (i=0 ; i<s->numedges ; i++) {
        e = dsurfedges[s->firstedge+i];
        if (e >= 0) v = dvertexes + dedges[e].v[0];
        else        v = dvertexes + dedges[-e].v[1];

        for (j=0 ; j<2 ; j++) {
 			// This is world->tex with world offset...
            val = (v->point[0] + faceoffset[0]) * tex->vecs[j][0]
                + (v->point[1] + faceoffset[1]) * tex->vecs[j][1]
                + (v->point[2] + faceoffset[2]) * tex->vecs[j][2]
                + tex->vecs[j][3];
            if (val < mins[j]) mins[j] = val;
            if (val > maxs[j]) maxs[j] = val;
        }
    }

    for (i=0 ; i<2 ; i++) {
        l->exactmins[i] = mins[i];
        l->exactmaxs[i] = maxs[i];

        mins[i] = floor(mins[i]/16);
        maxs[i] = ceil(maxs[i]/16);

        l->texmins[i] = mins[i];
        l->texsize[i] = maxs[i] - mins[i];
 
      if (l->texsize[i] > 17)
      		Error ("Bad surface extents");
    }
}

/*
 * =================
 * CalcPoints
 * For each texture aligned grid point, back project onto the plane
 * to get the world xyz value of the sample point
 * =================
 */
static int c_bad; // FIXME - what is this??
void CalcPoints (lightinfo_t *l)
{
    int    i;
    int    s, t;
    int    w, h, step;
    vec_t  starts, startt, us, ut;
    vec_t  *surf;
    vec_t  mids, midt;
    vec3_t facemid, move;

    /* fill in surforg                                         */
    /* the points are biased towards the center of the surface */
    /* to help avoid edge cases just inside walls              */

    surf = l->surfpt[0];
    mids = (l->exactmaxs[0] + l->exactmins[0])/2;
    midt = (l->exactmaxs[1] + l->exactmins[1])/2;

    tex_to_world(mids, midt, l, facemid);

    if (extrasamples) {    /* extra filtering */
        h = (l->texsize[1]+1)*2;
        w = (l->texsize[0]+1)*2;
        starts = (l->texmins[0]-0.5)*16;
        startt = (l->texmins[1]-0.5)*16;
        step = 8;
    } else {
        h = l->texsize[1]+1;
        w = l->texsize[0]+1;
        starts = l->texmins[0]*16;
        startt = l->texmins[1]*16;
        step = 16;
    }

    l->numsurfpt = w * h;
    for (t=0 ; t<h ; t++) {
        for (s=0 ; s<w ; s++, surf+=3) {
            us = starts + s*step;
            ut = startt + t*step;

            /* if a line can be traced from surf to facemid, point is good */
            for (i=0 ; i<6 ; i++) {
				tex_to_world(us, ut, l, surf);

                if (CastRay (facemid, surf) != -1)
                    break;    /* got it */
                if (i & 1) {  // i is odd
                    if (us > mids) {
                        us -= 8;
                        if (us < mids)
                            us = mids;
                    } else {
                        us += 8;
                        if (us > mids)
                            us = mids;
                    }
                } else {
                    if (ut > midt) {
                        ut -= 8;
                        if (ut < midt)
                            ut = midt;
                    } else {
                        ut += 8;
                        if (ut > midt)
                            ut = midt;
                    }
                }

                /* move surf 8 pixels towards the center */
                VectorSubtract (facemid, surf, move);
                VectorNormalize (move);
                VectorMA (surf, 8, move, surf);
            }
            if (i == 2)
                c_bad++;
        }
    }
}


/*
 * ============================================================================
 * FACE LIGHTING
 * ============================================================================
 */

static int c_culldistplane;
static int c_proper;

/*
 * ==============================================
 * TYRLITE: Attenuation formulae setup functions
 * ==============================================
 */
static vec_t scaledDistance(vec_t distance, const entity_t *light)
{
    switch (light->formula) {
    case 1:
    case 2:
    case 3:
        /* Return a small distance to prevent culling these lights, since we */
        /* know these formulae won't fade to nothing.                        */
        return (distance <= 0.0) ? -0.25 : 0.25;
    case 0:
        return scaledist * light->atten * distance;
    default:
        return 1;  /* shut up compiler warnings */
    }
}

static vec_t scaledLight(vec_t distance, const entity_t *light)
{
    const vec_t tmp = scaledist * light->atten * distance;
    switch (light->formula) {
    case 3: return light->light;
    case 1: return light->light / (tmp / 128);
    case 2: return light->light / ((tmp * tmp) / 16384);
    case 0:
        if (light->light > 0)
            return (light->light - tmp > 0) ? light->light - tmp : 0;
        else
            return (light->light + tmp < 0) ? light->light + tmp : 0;
    default:
        return 1;  /* shut up compiler warnings */
    }
}


/*
 * ================
 * SingleLightFace
 * ================
 */
static void SingleLightFace (const entity_t *light, lightinfo_t *l, const vec3_t faceoffset)
{
    vec_t    	dist;
    vec3_t   	incoming;
    vec_t    	angle;
    vec_t    	add;
    vec_t    	*surf;
    qboolean 	hit;
    int      	mapnum;
    int      	size;
    int     	c, i;
    vec3_t  	rel;
    vec3_t  	spotvec;
    vec_t   	falloff;
    vec_t   	*lightsamp;
    /* Colored lighting */
    vec3_t  	*lightcoloursamp;
    
    // pOx - START - Noise from Open Quartz
    int     	x, y, width; // SerpLord - noise coordinates
	vec_t   	*noise_surf; // SerpLord - used to "undersample" noise
	vec3_t  	point;       // current surface point snapped to grid
	float   	intensity=0; // noise intensity for the current point
	int     	seed;        // unique random seed for each light source
	// pOx - END - Noise from Open Quartz
	
	vec_t 		cast;
	
	faceoffset;// Shutup compiler warning
	
// DEBUG
//    qboolean face_done;

    VectorSubtract (light->origin, bsp_origin, rel);
    dist = scaledDistance((DotProduct(rel, l->facenormal) - l->facedist), light);

    /* don't bother with lights behind the surface */
    if (dist <= 0) return;

    /* don't bother with light too far away */
    if (dist > abs(light->light)) {
        c_culldistplane++;
        return;
    }

    if (light->targetent) {
        VectorSubtract (light->targetent->origin, light->origin, spotvec);
        VectorNormalize (spotvec);
        if (!light->angle)
            falloff = -cos(20*Q_PI/180);
        else
            falloff = -cos(light->angle/2*Q_PI/180);
    } else if (light->use_mangle) {
        VectorCopy (light->mangle, spotvec);
        if (!light->angle)
            falloff = -cos(20*Q_PI/180);
        else
            falloff = -cos(light->angle/2*Q_PI/180);
    } else {
        falloff = 0;    /* shut up compiler warnings */
    }

    mapnum = 0;
    for (mapnum=0 ; mapnum < l->numlightstyles ; mapnum++)
        if (l->lightstyles[mapnum] == light->style)
            break;

    lightsamp = l->lightmaps[mapnum];
    lightcoloursamp = l->lightmapcolours[mapnum];

    /* TYR: This way of doing things will generate "Too many light
     *      styles on a face" errors that aren't necessarily true.
     *      This is especially likely to happen if you make styled
     *      lights with delay != 0 as they are seen to be within
     *      lighting distance, even though they may be on totally
     *      opposite sides of the map (or separated by solid wall(s))
     *
     *      We should only generate the error when if find out that the
     *      face actually gets hit by the light
     *
     *      I think maybe MAXLIGHTMAPS is actually supposed to be 3.
     *      Have to make some more test maps (or check the engine source)
     *      for this.
     */

    if (mapnum == l->numlightstyles) {    /* init a new light map */
//        if (mapnum == MAXLIGHTMAPS) {
//            logprint ("WARNING: Too many light styles on a face\n");
//            logprint ("   lightmap point near (%0.0f, %0.0f, %0.0f)\n",
//                      l->surfpt[0][0], l->surfpt[0][1], l->surfpt[0][2]);
//            logprint ("   light->origin (%0.0f, %0.0f, %0.0f)\n",
//                      light->origin[0], light->origin[1], light->origin[2]);
//            return;
//        }
        size = (l->texsize[1]+1)*(l->texsize[0]+1);
        for (i=0 ; i<size ; i++) {
            if (colored) {
                lightcoloursamp[i][0] = 0;
                lightcoloursamp[i][1] = 0;
                lightcoloursamp[i][2] = 0;
            }
            lightsamp[i] = 0;
        }
    }

    /* check it for real */

// DEBUG
//face_done = false;

    hit = false;
    c_proper++;

    surf = l->surfpt[0];
    for (c=0 ; c<l->numsurfpt ; c++, surf+=3)
    {
    	// pOx - optimization
		if ((cast = CastRay(light->origin, surf)) == -1) continue;   
       
		dist = scaledDistance(cast, light);
		if (dist < 0) continue;	/* light doesn't reach */
	
        VectorSubtract (light->origin, surf, incoming);
        VectorNormalize (incoming);        
        
#if TYRLIGHT_TEST_PHONG
	{
		vec3_t pointnormal = {0,0,0};
		GetPhongNormal(l->surfnum, surf, pointnormal);
		angle = DotProduct(incoming, pointnormal);
		if (!VectorCompare(pointnormal, l->facenormal)) modnorms++; else facenorms++;//DEBUG
	}
#else
        angle = DotProduct(incoming, l->facenormal);
#endif

        if (light->targetent || light->use_mangle) { /* spotlight cutoff */
            if (DotProduct (spotvec, incoming) > falloff)
                continue;
        }

        angle = (1.0-scalecos) + scalecos*angle;
        add = scaledLight(cast, light);
		
		// pOx - START - Noise from Open Quartz
		if (light->noise) {
			noise_surf = surf;
			if (extrasamples)
			{
				// We don't want to oversample noise because that just
				// waters it down.  So we "undersample" noise by using
				// the same surf coord for every group of 4 lightmap pixels
				// ("undersampling", "pixelation", "anti-interpolation" :-)
				width = (l->texsize[0] + 1) * 2;
				x = c % width;
				y = c / width;
				if (x % 2 && y % 2) noise_surf -= width * 3 + 3;
				else if (y % 2) noise_surf -= width * 3;
				else if (x % 2) noise_surf -= 3;
			}

			if (light->noisetype == NOISE_SMOOTH)
				snap_vector (noise_surf, point, 0);
			else
				snap_vector (noise_surf, point, light->resolution);

			seed = light - entities;

			if (light->noisetype == NOISE_RANDOM)
				intensity = noise3d (point, seed);
			if (light->noisetype == NOISE_SMOOTH)
				intensity = noise_scaled (point, light->resolution, seed);
			if (light->noisetype == NOISE_PERLIN)
				intensity = noise_perlin (point, light->persistence, seed);

			add *= intensity * light->noise + 1.0 - light->noise;
		}
		// pOx - END - Noise from Open Quartz
		
		add *= angle;
        lightsamp[c] += add;

        if (colored) {
            add /= (vec_t)255.0;
            lightcoloursamp[c][0] += add * light->lightcolour[0];
            lightcoloursamp[c][1] += add * light->lightcolour[1];
            lightcoloursamp[c][2] += add * light->lightcolour[2];
        }

        if (abs(lightsamp[c]) > 1) hit = true;/* ignore really tiny lights */   
    }

    if (mapnum == l->numlightstyles && hit) {
        if (mapnum == MAXLIGHTMAPS-1) {
            printf ("WARNING: Too many light styles on a face\n");
            printf ("   lightmap point near (%0.0f, %0.0f, %0.0f)\n",
                      l->surfpt[0][0], l->surfpt[0][1], l->surfpt[0][2]);
            printf ("   light->origin (%0.0f, %0.0f, %0.0f)\n",
                      light->origin[0], light->origin[1], light->origin[2]);
            return;
        }
/*
// DEBUG
if (mapnum == MAXLIGHTMAPS-1  && !face_done) {
    logprint ("DEBUG: MAXLIGHTMAPS on lightmap near (%0.0f, %0.0f, %0.0f)\n",
              l->surfpt[0][0], l->surfpt[0][1], l->surfpt[0][2]);
    logprint ("                       light->origin (%0.0f, %0.0f, %0.0f)\n",
              light->origin[0], light->origin[1], light->origin[2]);
    face_done = true;
}
*/
        l->lightstyles[mapnum] = light->style;
        l->numlightstyles++;    /* the style has some real data now */
    }
}

/*
 * =============
 * SkyLightFace
 * =============
 */
void SkyLightFace (lightinfo_t *l, const vec3_t faceoffset)
{
    int    i,j;
    vec_t  *surf;
    vec3_t incoming;
    vec_t  angle;
	
	faceoffset;// Shutup compiler warning
	
    /* Don't bother if surface facing away from sun */
    if (DotProduct (sunmangle, l->facenormal) <= 0)
        return;

    /* if sunlight is set, use a style 0 light map */
    for (i=0 ; i< l->numlightstyles ; i++)
        if (l->lightstyles[i] == 0)
            break;
    if (i == l->numlightstyles) {
        if (l->numlightstyles == MAXLIGHTMAPS)
            return;        /* oh well, too many lightmaps... */
        l->lightstyles[i] = 0;
        l->numlightstyles++;
    }

    /* Check each point... */
    VectorCopy(sunmangle, incoming);
    VectorNormalize(incoming);
    angle = DotProduct (incoming, l->facenormal);
    angle = (1.0-scalecos) + scalecos*angle;
    surf = l->surfpt[0];
    for (j=0 ; j<l->numsurfpt ; j++, surf+=3) {
        if (TestSky(surf, sunmangle)) {
            l->lightmaps[i][j] += (angle*sunlight);
            if (colored) {
                l->lightmapcolours[i][j][0] +=
                    (angle * sunlight * sunlight_color[0]) /255;
                l->lightmapcolours[i][j][1] +=
                    (angle * sunlight * sunlight_color[1]) /255;
                l->lightmapcolours[i][j][2] +=
                    (angle * sunlight * sunlight_color[2]) /255;
            }
        }
    }
}

/*
 * ============
 * FixMinlight
 * ============
 */
void FixMinlight (lightinfo_t *l)
{
    int   i, j, k;
    vec_t tmp;

    /* if minlight is set, there must be a style 0 light map */
    for (i=0 ; i< l->numlightstyles ; i++) {
		if (l->lightstyles[i] == 0) {
	    // Set the minimum light level...
            break;
		} else {
	    // Set zero minimum...
		}
    }
    if (i == l->numlightstyles) {
        if (l->numlightstyles == MAXLIGHTMAPS)
            return;        /* oh well.. */
        for (j=0 ; j<l->numsurfpt ; j++)
            l->lightmaps[i][j] = worldminlight;
        if (colored)
            for (j=0 ; j<l->numsurfpt ; j++) {
                l->lightmapcolours[i][j][0] =
                    (worldminlight * minlight_color[0]) / (vec_t)255;
                l->lightmapcolours[i][j][1] =
                    (worldminlight * minlight_color[1]) / (vec_t)255;
                l->lightmapcolours[i][j][2] =
                    (worldminlight * minlight_color[2]) / (vec_t)255;
            }
        l->lightstyles[i] = 0;
        l->numlightstyles++;
    } else {
        for (j=0 ; j<l->numsurfpt ; j++) {
            if ( l->lightmaps[i][j] < worldminlight)
                l->lightmaps[i][j] = worldminlight;
            if (colored) {
                for (k=0 ; k<3 ; k++) {
                    tmp = (vec_t)worldminlight * minlight_color[k];
		    		tmp /= (vec_t)255.0;
                    if (l->lightmapcolours[i][j][k] < tmp )
                        l->lightmapcolours[i][j][k] = tmp;
                }
            }
        }
    }
}


/*
 * light is the light intensity, needed to check if +ve or -ve.
 * src and dest are the source and destination color vectors (vec3_t).
 * dest becomes a copy of src where
 *    MakePosColoredLight zeros negative light components.
 *    MakeNegColoredLight zeros positive light components.
 */
static void MakePosColoredLight(int light, vec3_t dest, vec3_t src)
{
    int i;
    if (light >= 0) {
	for (i=0 ; i<3 ; i++)
	    if (src[i] < 0) dest[i] = 0;
	    else            dest[i] = src[i];
    } else {
	for (i=0 ; i<3 ; i++)
	    if (src[i] > 0) dest[i] = 0;
	    else            dest[i] = src[i];
    }
}

static void MakeNegColoredLight(int light, vec3_t dest, const vec3_t src) {
    int i;
    if (light >= 0) {
	for (i=0 ; i<3 ; i++)
	    if (src[i] > 0) dest[i] = 0;
	    else            dest[i] = src[i];
    } else {
	for (i=0 ; i<3 ; i++)
	    if (src[i] < 0) dest[i] = 0;
	    else            dest[i] = src[i];
    }
}

/*
 * ============
 * LightFace
 * ============
 */
void LightFace (int surfnum, qboolean nolight, const vec3_t faceoffset)
{
    dface_t     *f;
    /*static*/// pOx - (CW 6<)
   	lightinfo_t l;
    int         s, t;
    int         i,j,k,c;

    /* TYR - temp vars */
    vec_t       max;
    int         x1,x2,x3,x4;

    vec_t       total;
    int         size;
    int         lightmapwidth;
    int         lightmapsize;
    byte        *out;
    byte        *lit_out;
    vec_t       *light;

    /* TYR - colored lights */
    vec3_t      *lightcolour;
    vec3_t      totalcolours;

    int         w, h;
    vec3_t      point;
	
	nolight; faceoffset;// Shutup compiler warning
	
    f = dfaces + surfnum;

    /* some surfaces don't need lightmaps */
    f->lightofs = -1;
    for (j=0 ; j<MAXLIGHTMAPS ; j++)
        f->styles[j] = 255;

    if ( texinfo[f->texinfo].flags & TEX_SPECIAL)
        return;  /* non-lit texture */

    memset (&l, 0, sizeof(l));
    l.surfnum = surfnum;
    l.face = f;

    /* rotate plane */

	VectorCopy (dplanes[f->planenum].normal, l.facenormal);

    l.facedist = dplanes[f->planenum].dist;
    VectorScale (l.facenormal, l.facedist, point);
    VectorAdd( point, faceoffset, point );
    l.facedist = DotProduct( point, l.facenormal );

    if (f->side) {
        VectorSubtract (vec3_origin, l.facenormal, l.facenormal);
        l.facedist = -l.facedist;
    }

    CalcFaceVectors (&l);
    CalcFaceExtents (&l, faceoffset);
    CalcPoints (&l);

    lightmapwidth = l.texsize[0]+1;
    size = lightmapwidth*(l.texsize[1]+1);
    
    if (size > SINGLEMAP)
        Error ("Bad lightmap size");

    for (i=0 ; i<MAXLIGHTMAPS ; i++)
        l.lightstyles[i] = 255;

    /* Under normal circumstances, the lighting procedure is:
     * - cast all light entities
     * - cast sky lighting
     * - do minlighting.
     *
     * However, if nominlimit is enabled then we need to do the following:
     * - cast _positive_ lights
     * - cast _positive_ skylight (if any)
     * - do minlighting
     * - cast _negative_ lights
     * - cast _negative_ sky light (if any)
     */

    l.numlightstyles = 0;
    if (nominlimit) {
        /* cast only positive lights */
        entity_t lt;

        for (i=0 ; i<num_entities ; i++) {	    	
	    	
	    	if(entities[i].light && entities[i].notents && !worldface[surfnum])
	    		continue;// pOx - this light doesn't affect sub models
	    	
	    	if (colored) {
				if (entities[i].light) {
		    		memcpy(&lt, &entities[i], sizeof(entity_t));
		    		MakePosColoredLight (entities[i].light, (vec_t *)lt.lightcolour, (vec_t *)entities[i].lightcolour);
		    		SingleLightFace (&lt, &l, faceoffset);
				}
	    	} else if (entities[i].light > 0) {
				SingleLightFace (&entities[i], &l, faceoffset);
	    	}
        }
        /* cast positive sky light */
        if (sunlight) {
	    	if (colored) {
				vec3_t suncol_save;

				VectorCopy (sunlight_color, suncol_save);
				MakePosColoredLight (sunlight, sunlight_color, suncol_save);
				SkyLightFace (&l, faceoffset);
				VectorCopy (suncol_save, sunlight_color);
	    	} else if (sunlight > 0) {
				SkyLightFace (&l, faceoffset);
	    	}
        }
    } else {
        /* (!nominlimit) => cast all lights */
        for (i=0 ; i<num_entities ; i++) {
        	
        	if(entities[i].light && entities[i].notents && !worldface[surfnum])
	    		continue;// pOx - this light doesn't affect sub models
	    		
            if (entities[i].light) SingleLightFace (&entities[i], &l, faceoffset);
		}
		
        /* cast sky light */
        if (sunlight) SkyLightFace (&l, faceoffset);
    }

    /* Minimum lighting */
    FixMinlight (&l);

    if (nominlimit) {
        /* cast only negative lights */
        entity_t lt;

        for (i=0 ; i<num_entities ; i++) {
        	
			if(entities[i].light && entities[i].notents && !worldface[surfnum])
	    		continue;// pOx - this light doesn't affect sub models
        	
	    	if (colored) {
				if (entities[i].light) {
		    		memcpy(&lt, &entities[i], sizeof(entity_t));
		    		MakeNegColoredLight (entities[i].light, (vec_t *)lt.lightcolour, (vec_t *)entities[i].lightcolour);
					SingleLightFace (&lt, &l, faceoffset);
				}
	    	} else if (entities[i].light < 0) {
				SingleLightFace (&entities[i], &l, faceoffset);
	    	}
        }
        /* cast negative sky light */
        if (sunlight) {
	    	if (colored) {
				vec3_t suncol_save;

				VectorCopy (sunlight_color, suncol_save);
				MakeNegColoredLight (sunlight, sunlight_color, suncol_save);
				SkyLightFace (&l, faceoffset);
				VectorCopy (suncol_save, sunlight_color);
			} else if (sunlight < 0) {
				SkyLightFace (&l, faceoffset);
	    	}
		}

        /* Fix any negative values */
		for (i=0; i < l.numlightstyles; i++) {
	    	for (j=0; j < l.numsurfpt; j++) {
				if (l.lightmaps[i][j] < 0) {
		    		l.lightmaps[i][j] = 0;
				}
				if (colored) {
		    		for (k=0; k<3; k++) {
						if (l.lightmapcolours[i][j][k] < 0) {
			    			l.lightmapcolours[i][j][k] = 0;
						}
		    		}
				}
	    	}
		}
    }

    if (!l.numlightstyles) return;    /* no light hitting it */

    /* save out the values */

    for (i=0 ; i <MAXLIGHTMAPS ; i++)
        f->styles[i] = l.lightstyles[i];

    /* Extra room for BSP30 lightmaps */
    if (colored && !litfile)
		lightmapsize = size*l.numlightstyles*color_bytes;// pOx - RGBA | RGB
    else
		lightmapsize = size*l.numlightstyles;

    out = GetFileSpace (lightmapsize);
 
    if (litfile) lit_out = (byte *)GetLitFileSpace (lightmapsize*3); else lit_out = NULL;
	
    f->lightofs = out - filebase;
    
    /* extra filtering */
    h = (l.texsize[1]+1)*2;
    w = (l.texsize[0]+1)*2;

    for (i=0 ; i< l.numlightstyles ; i++) {
        if (l.lightstyles[i] == 0xff)
            Error ("Wrote empty lightmap");

        light = l.lightmaps[i];
        lightcolour = l.lightmapcolours[i];
        c = 0;

        for (t=0 ; t<=l.texsize[1] ; t++) {
            for (s=0 ; s<=l.texsize[0] ; s++, c++) {
                if (extrasamples) {
                    x1 = t*2*w+s*2;
                    x2 = x1+1;
                    x3 = (t*2+1)*w+s*2;
                    x4 = x3+1;

                    /* filtered sample */
                    total = light[x1] + light[x2] + light[x3] + light[x4];
                    total *= 0.25;

                    /* Calculate the colour */
                    if (colored) {
                        totalcolours[0] =
                            lightcolour[x1][0] + lightcolour[x2][0]
                            + lightcolour[x3][0] + lightcolour[x4][0];
                        totalcolours[0] *= 0.25;
                        totalcolours[1] =
                            lightcolour[x1][1] + lightcolour[x2][1]
                            + lightcolour[x3][1] + lightcolour[x4][1];
                        totalcolours[1] *= 0.25;
                        totalcolours[2] =
                            lightcolour[x1][2] + lightcolour[x2][2]
                            + lightcolour[x3][2] + lightcolour[x4][2];
                        totalcolours[2] *= 0.25;
                    }
                } else {
                    total = light[c];
                    if (colored) VectorCopy (lightcolour[c], totalcolours);
                }
                total *= rangescale;    /* scale before clamping */

                /* CSL - Scale back intensity, instead of capping individual */
                /*       colours                                             */
                if (colored) {
                    VectorScale (totalcolours, rangescale, totalcolours);
                    max = 0.0;
                    for (j = 0; j < 3; j++) {
                        if (totalcolours[j] > max) {
                            max = totalcolours[j];
                        } else if (totalcolours[j] < 0.0f) {
                            Error ("colour %i < 0", j);
                        }
                    }
                    if (max > 255.0f)
                        VectorScale (totalcolours, 255.0f / max, totalcolours);
                }

                if (total > 255.0f)
                    total = 255.0f;
                 else if (total < 0)
		    		total = 0;

				/* Write out the lightmap in the appropriate format */
				if (colored && !litfile) {
					*out++ = totalcolours[0];
					*out++ = totalcolours[1];
					*out++ = totalcolours[2];
				}
				if (colored && litfile) {
				    *lit_out++ = totalcolours[0];
				    *lit_out++ = totalcolours[1];
				    *lit_out++ = totalcolours[2];
				}
				
				if (bspversion != BSP_EQUAKE_VERSION)// pOx - not in RGB format
		            *out++ = total;
		
			}
		}

	}
}
